package mySynchMeans;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.CyclicBarrier;
//import java.util.concurrent.ExecutorService;
//import java.util.concurrent.Executors;
import java.util.concurrent.atomic.AtomicInteger;
import org.junit.AfterClass;
import static org.junit.Assert.*;
import org.junit.BeforeClass;
import org.junit.Test;



public class DormitorySemSemTest {
    //private static final ExecutorService pool = Executors.newCachedThreadPool();

    public DormitorySemSemTest() {
    }

    @BeforeClass
    public static void setUpClass() {
    }

    @AfterClass
    public static void tearDownClass() {
    }

    
    /**
     * Test of toSleep method, of class DormitorySemSem.
     * In this test we make n threads invoke toSleep() on an n-seat dormitory, 
     * and we check that all of them actually block.
     */
    @Test(timeout=1000)
    public void testBlockingForToSleep(){
        final int NUM_SEATS = 5;
        int NUM_THREADS = NUM_SEATS;
        final Dormitory dorm = new DormitorySemSem(NUM_SEATS);
        dorm.setVerbose(false);
        final CyclicBarrier barrier = new CyclicBarrier(NUM_THREADS);
        int TIMEOUT = 100; //milliseconds
        final AtomicInteger failedInBlocking = new AtomicInteger(0);
        List<Thread> threads = new ArrayList<>();

        System.out.println("TestRunner - Test toSleep() - with n places and n threads, all treads block");

        class Worker extends Thread{
            @Override
            public void run(){
                try {
                    barrier.await(); // all the threads will start together
                
                    dorm.toSleep();
                    
                    failedInBlocking.incrementAndGet(); 
                    
                    // in this test, no thread should reach this point
                
                    barrier.await(); // all the threads would finish together
                    
                }  catch (BrokenBarrierException | InterruptedException e) {}
            }
        }
            
        Worker w;
        for (int i=0; i<NUM_THREADS; i++) {
            w = new Worker();
            threads.add(w);
            w.start();
        }
        
        try {
            Thread.sleep(TIMEOUT); // here we wait user to block on toSleep()
        } catch (InterruptedException ex) {}

        for(Thread t: threads){
            t.interrupt();  //this interrupt will wake up threads inside the dormitory
        }
        
        assertEquals("Some threads didn't block!", 0 ,failedInBlocking.get());
        
    }
    

    
    /**
     * Test of toSleep/wakeUpAll methods, of class DormitorySemSem.
     * In this test we make n+1 threads invoke toSleep() on an n-seat dormitory, 
     * and we check that, after invoking wakeUpAll(), only one of them 
     * is kept blocked.
     */
    @Test(timeout=1000)
    public void testBlockingTheRest(){
        int numSeats = 5;
        int numThreads = numSeats+1;
        final Dormitory dorm = new DormitorySemSem(numSeats);
        dorm.setVerbose(false);
        final CyclicBarrier barrier = new CyclicBarrier(numThreads);
        int TIMEOUT = 100; //milliseconds
        final AtomicInteger passedThreads = new AtomicInteger(0);
        final AtomicInteger onGateThreads = new AtomicInteger(0);
        List<Thread> threads = new ArrayList<>();

        System.out.println("TestRunner - Test toSleep()/wakeUpAll() - with n places and n+1 threads, "+
                "after wakeUpAll(), only one thread does not proceed");

        class Worker extends Thread{
            @Override
            public void run(){
                try {
                    barrier.await(); // all the threads will start together
                    
                    try {
                        dorm.toSleep();
                    } catch (InterruptedException e) {
                        onGateThreads.incrementAndGet();
                        System.out.println("    !!! Thread interrupted while waiting before the gate."); 
                        throw(e);}
                    
                    passedThreads.incrementAndGet();   
                
                    barrier.await(); // all the threads will finish together
                    
                }  catch (BrokenBarrierException | InterruptedException e) {}
            }
        }
            
        Worker w;
        for (int i=0; i<numThreads; i++) {
            w=new Worker();
            threads.add(w);
            w.start();
        }
        
        try {
            Thread.sleep(TIMEOUT);
        
            dorm.wakeUpAll();
            
            Thread.sleep(TIMEOUT); 
        } catch (InterruptedException ex) {}

        for(Thread t: threads){
            t.interrupt();
        }
        
        try {
            Thread.sleep(TIMEOUT); //to make sure the other threads have finished 
                                   //reacting to the received interrupt
            
            if(passedThreads.get()!=numSeats) { // same as: assertEquals(numSeats,passedThreads.get()); 
                fail();
            }
            if(onGateThreads.get()!=numThreads-numSeats) { // ditto
                System.out.println("onGateThreads: "+onGateThreads.get());
                fail();
            }
        } catch (InterruptedException e) {}
    }
    
    
    
    
/**
     * Test of consistent use of toSleep/wakeUpAll methods, of class DormitorySemSem.
     * In this test we make n=p+s threads repeatedly (m times) invoke toSleep() 
     * on an p-seat dormitory, while the main thread periodically invokes 
     * ceil(n*m/p) times wakeUpAll() over a (presumably) full dormitory.
     * At the end, no thread should be blocked.
     * PAY ATTENTION: results depend on timings!!!!
     */
    @Test//(timeout=10000)   // ! TIMEOUT > NUMWAKEUPS*WU_PERIOD+epsilon
    public void testWakeUpAllBlocked(){
        int NUM_SEATS = 4;
        int SURPLUS = 3;
        final int NUM_THREADS = NUM_SEATS+SURPLUS;
        final int NUM_CYCLES = 100;
        final int NUM_WAKEUPS = NUM_CYCLES*NUM_THREADS/NUM_SEATS + ((NUM_CYCLES*NUM_THREADS)%NUM_SEATS == 0 ? 0:1) ;
        final int MAX_CYCLE_DELAY = 10; //milliseconds
        final DormitorySemSem dorm = new DormitorySemSem(NUM_SEATS, false);
        final CyclicBarrier barrier = new CyclicBarrier(NUM_THREADS);
        final int WU_PERIOD = 40; //milliseconds
        final AtomicInteger totalTransits = new AtomicInteger(0);
        List<Thread> threads = new ArrayList<>();

        System.out.println("TestRunner - Torture test on toSleep()/wakeUpAll() - "+
                "with p places and t=p+s threads, after repeatedly invoking wakeUpAll(), "+
                "all invocations to toSleep() terminate.");

        
        System.out.println("threads: "+NUM_THREADS + " num_wu: "+NUM_WAKEUPS);
        class Worker extends Thread{
            private final java.util.Random myRNG = new java.util.Random();
            public int myTransits = 0; //not acyually used in the test
            
            @Override
            public void run(){
                try {
                    barrier.await();
                    
                    for(int i=1; i<=NUM_CYCLES; i++) {
                        
                        Thread.sleep(myRNG.nextInt(MAX_CYCLE_DELAY));
                
                        dorm.toSleep();
                        
                        totalTransits.incrementAndGet();
                        myTransits++;
                    }
                
                    barrier.await();
                    
                }  catch (BrokenBarrierException | InterruptedException e) {}
            }
        }
            
        Worker w;
        for (int i=0; i<NUM_THREADS; i++) {
            w=new Worker();
            threads.add(w);
            w.start();
        }
        
        try {
            for(int i=1; i<=NUM_WAKEUPS; i++){
                Thread.sleep(WU_PERIOD);
                
                //assertEquals(NUM_SEATS, dorm.wakeUpAll());
                int howmany = dorm.wakeUpAll();

                if (howmany < Math.min(NUM_SEATS, NUM_THREADS) ) {
                    System.out.println("At wakeup # "+i+", waken "+ howmany+ 
                            " from "+ NUM_SEATS + "seats and # of transits: "+ totalTransits.get());
                    System.out.println("Access permits # "+dorm.accessGate.availablePermits()+", waken "+ howmany+ 
                            " sleeping permits # "+ dorm.sleepingPlaces.availablePermits() + 
                            " and occupied seats # "+ dorm.occupiedSeats.get());
                }
            }
        
        } catch (InterruptedException ex) {}
        
        
        //the following part tries to re-catch threads that possibly got late
        //in executing the last calls of toSleep()
        int nretries=10;
        int howmanyblocked;
        
        do {
            howmanyblocked = (barrier.getParties()-barrier.getNumberWaiting())%barrier.getParties();
            if (howmanyblocked>0) {
                try {
                    Thread.sleep(WU_PERIOD);
                    dorm.wakeUpAll();
                } catch (InterruptedException ex) {}
            }
            else break;
            nretries--;
        } while(nretries>0);
        
        
        howmanyblocked = (barrier.getParties()-barrier.getNumberWaiting())%barrier.getParties();
        System.out.println("BLOCKED:"+howmanyblocked);
        

        for(Thread t: threads){
            t.interrupt();
        }
        
        if(totalTransits.get()!= NUM_THREADS*NUM_CYCLES) {
            System.out.println("#blocked: "+howmanyblocked+" #transits: "
                    +totalTransits.get()+" on "+ NUM_THREADS*NUM_CYCLES);
            System.out.println("acc. perms: "+dorm.accessGate.availablePermits()+
                    " sleeping perms: "
                    +dorm.sleepingPlaces.availablePermits() );
            
        assertEquals(NUM_THREADS*NUM_CYCLES, totalTransits.get());
        
        }
        
    }
    
    

}
